"""
Types of Errors specified by personio-py
"""
from typing import Any, Dict, Type

from requests import Response


class PersonioError(Exception):
    """A generic error caused by personio-py"""


class MissingCredentialsError(PersonioError):
    """you are missing some strictly required credentials"""


class PersonioApiError(PersonioError):
    """
    An error response from the Personio HTTP API

    :param status_code: the HTTP status code of the API response
    :param message: the error message from the Personio API
    :param error_code: the error code that is provided with the Personio API response
           (not the HTTP status code!)
    :param errors: details about the errors that are provided by the Personio API as dictionary
    :param response: the HTTP response
    """

    def __init__(self, status_code: int, message: str, error_code: int = None,
                 errors: Dict[str, Any] = None, response: Response = None):
        super().__init__()
        self.status_code = status_code
        self.message = message
        self.error_code = error_code
        self.errors = errors
        self.response = response

    @classmethod
    def from_response(cls, response: Response):
        """
        Creates a ``PersonioApiError`` from the specified HTTP response.

        :param response: a HTTP error response from Personio
        :return: a PersonioApiError that matches the HTTP error
        """
        try:
            data: Dict = response.json()
            error = data.get('error', {})
            code = error.get('code')
            message = error.get('message')
            error_dict: Dict = error.get('errors')
            return PersonioApiError(
                status_code=response.status_code,
                message=message,
                error_code=code,
                errors=error_dict,
                response=response)
        except ValueError:
            return PersonioApiError(
                status_code=response.status_code,
                message=response.text,
                response=response)

    def __str__(self):
        message = f"request failed with HTTP status code {self.status_code}: {self.message}"
        if self.error_code:
            message += f" (error code {self.error_code})"
        if self.errors:
            if message[-1] != '.':
                message += '.'
            message += f" Details: {self.errors}"
        return message


class UnsupportedMethodError(PersonioError):
    """this method is not supported by this class (but it might be by a similar one)"""

    def __init__(self, method: str, clazz: Type):
        super().__init__()
        self.method = method
        self.clazz = clazz

    def __str__(self):
        return f"method '{self.method}' is not available for {self.clazz.__name__}"



"""
mappings from Personio API fields to Python data types and vice versa are defined in this module
"""
import logging
import re
from datetime import date, datetime, timedelta
from decimal import Decimal
from typing import Any, Dict, List, NamedTuple, Optional, TYPE_CHECKING, Type, TypeVar, Union

if TYPE_CHECKING:
    from personio_py import Personio
    from personio_py.models import PersonioResourceType

logger = logging.getLogger('personio_py')

T = TypeVar('T')


class FieldMapping:
    """
    A generic mapping from a Personio API field to a Python object.
    The default implementation works great for strings, but for more complex types,
    please refer to the subclasses of ``FieldMapping``.

    :param api_field: name of the field in the Personio API
    :param class_field: name of the attribute in the target Python object
    :param field_type: data type of the field
    """

    def __init__(self, api_field: str, class_field: str, field_type: Type[T]):
        self.api_field = api_field
        self.class_field = class_field
        self.field_type = field_type

    def serialize(self, value: T) -> Union[str, Dict]:
        """
        Serialize the parsed value to the format that the Personio API expects.

        :param value: the value to serialize
        :return: the serialized value
        """
        return str(value)

    def deserialize(self, value: Union[str, Dict], **kwargs) -> T:
        """
        Deserialize the Personio API value to a more useful Python data type.

        :param value: the value as provided by the Personio API
        :param kwargs: additional parameters (to be used by subclasses)
        :return: the deserialized value
        """
        return self.field_type(value)

    def __str__(self):
        return f"{self.__class__.__name__} {self.__dict__}"


class NumericFieldMapping(FieldMapping):
    # don't touch numeric types, unless they are strings...

    def __init__(self, api_field: str, class_field: str, field_type=float):
        super().__init__(api_field, class_field, field_type=field_type)

    def serialize(self, value: Union[int, float, str]) -> Union[int, float, str]:
        return value

    def deserialize(self, value: Union[int, float, str], **kwargs) -> Union[int, float, str]:
        return self.field_type(value) if isinstance(value, str) else value


class BooleanFieldMapping(FieldMapping):

    def __init__(self, api_field: str, class_field: str):
        super().__init__(api_field, class_field, field_type=bool)

    def serialize(self, value: bool) -> bool:
        return value

    def deserialize(self, value: str, **kwargs) -> bool:
        return bool(value)


class DateTimeFieldMapping(FieldMapping):

    def __init__(self, api_field: str, class_field: str):
        super().__init__(api_field, class_field, field_type=datetime)

    def serialize(self, value: datetime) -> str:
        return value.isoformat()

    def deserialize(self, value: str, **kwargs) -> datetime:
        return datetime.fromisoformat(value)


class DateFieldMapping(FieldMapping):

    def __init__(self, api_field: str, class_field: str):
        super().__init__(api_field, class_field, field_type=date)

    def serialize(self, value: datetime) -> str:
        return value.isoformat()

    def deserialize(self, value: str, **kwargs) -> date:
        return date.fromisoformat(value[:10])


class DurationFieldMapping(FieldMapping):

    pattern = re.compile(r"\d\d?:\d\d")

    def __init__(self, api_field: str, class_field: str):
        super().__init__(api_field, class_field, field_type=timedelta)

    def serialize(self, value: timedelta) -> str:
        mm, ss = divmod(value.total_seconds(), 60)
        hh, mm = divmod(mm, 60)
        return f"{int(hh):02d}:{int(mm):02d}"

    def deserialize(self, value: str, **kwargs) -> timedelta:
        return self.str_to_timedelta(value)

    @classmethod
    def str_to_timedelta(cls, s: str) -> timedelta:
        if not isinstance(s, str):
            raise TypeError(f"expected a string, but got {type(s)}")
        trimmed = s.strip()
        if cls.pattern.fullmatch(trimmed):
            hh, mm = trimmed.split(':')
            return timedelta(hours=int(hh), minutes=int(mm))
        else:
            raise ValueError(f"the string '{s}' does not represent a valid duration. "
                             f"Expected format is 'hh:mm', e.g. '06:30'.")


class MultiTagFieldMapping(FieldMapping):

    def __init__(self, api_field: str, class_field: str):
        super().__init__(api_field, class_field, field_type=list)

    def serialize(self, values: List[str]) -> str:
        for value in values:
            if ',' in value:
                raise ValueError(
                    f"Due to a restrictions at Personio, no commas are allowed in "
                    f"multi selection fields, please adjust '{value}'")
        return ','.join(values)

    def deserialize(self, value: str, **kwargs) -> List[str]:
        return [s.strip() for s in value.split(',')] if value else []


class ObjectFieldMapping(FieldMapping):

    def __init__(self, api_field: str, class_field: str, field_type: Type['PersonioResourceType']):
        super().__init__(api_field, class_field, field_type)

    def serialize(self, value: 'PersonioResourceType') -> Dict:
        if self.field_type._flat_dict:
            return value.to_dict(nested=True)
        else:
            return {
                'type': self.field_type._api_type_name,
                'attributes': value.to_dict(nested=True)
            }

    def deserialize(self, value: Dict, client: 'Personio' = None) \
            -> Optional['PersonioResourceType']:
        if value and isinstance(value, dict):
            if not self.field_type._flat_dict:
                value = value['attributes']
            return self.field_type.from_dict(value, client=client)
        else:
            return None


class ListFieldMapping(FieldMapping):
    # wraps another field mapping, to handle list types
    # e.g. ``ListFieldMapping(ObjectFieldMapping('cost_centers', 'cost_centers', CostCenter))``

    def __init__(self, item_mapping: FieldMapping):
        super().__init__(item_mapping.api_field, item_mapping.class_field, field_type=List)
        self.item_mapping = item_mapping

    def serialize(self, values: List[Any]) -> List[Any]:
        return [self.item_mapping.serialize(item) for item in values]

    def deserialize(self, values: List[Any], client: 'Personio' = None) -> List[Any]:
        return [self.item_mapping.deserialize(item, client=client) for item in values]


FieldMappingType = TypeVar('FieldMappingType', bound=FieldMapping)


class DynamicMapping(NamedTuple):
    """
    Defines a mapping from a dynamic field to a more memorable name and its actual data type,
    so that it may be converted into a proper python type, if possible.
    """
    field_id: int
    """the id number of the dynamic field, e.g. for 'dynamic_123456', field_id=123456"""
    alias: str
    """a more memorable name than the field_id, will be used as dictionary key"""
    data_type: Type[T]
    """the data type of the field, for automatic conversion (e.g. str to datetime)"""

    def get_field_mapping(self) -> FieldMappingType:
        api_field = f'dynamic_{self.field_id}'
        if self.data_type == str:
            return FieldMapping(api_field, self.alias, str)
        elif self.data_type in (int, float, Decimal):
            return NumericFieldMapping(api_field, self.alias, self.data_type)
        elif self.data_type == date:
            return DateFieldMapping(api_field, self.alias)
        elif self.data_type == datetime:
            return DateTimeFieldMapping(api_field, self.alias)
        elif self.data_type == timedelta:
            return DurationFieldMapping(api_field, self.alias)
        elif self.data_type in (list, List):
            return MultiTagFieldMapping(api_field, self.alias)
        else:
            logger.warning(f"unexpected type {self.data_type} for dynamic field {self.field_id}")
            return FieldMapping(api_field, self.alias, self.data_type)



"""
Definition of ORMs for objects that are available in the Personio API
"""
import json
import logging
from collections import namedtuple
from datetime import datetime, timedelta
from functools import total_ordering
from typing import Any, Dict, List, NamedTuple, Optional, TYPE_CHECKING, Tuple, Type, TypeVar

from personio_py import PersonioError, UnsupportedMethodError
from personio_py.mapping import BooleanFieldMapping, DateFieldMapping, DateTimeFieldMapping, \
    DurationFieldMapping, DynamicMapping, FieldMapping, ListFieldMapping, NumericFieldMapping, \
    ObjectFieldMapping

if TYPE_CHECKING:
    # only type checkers may import Personio, otherwise we get an evil circular import error
    from personio_py import Personio

logger = logging.getLogger('personio_py')


class DynamicAttr(NamedTuple):
    field_id: int
    label: str
    value: Any

    @classmethod
    def from_attributes(cls, d: Dict[str, Dict[str, Any]]) -> List['DynamicAttr']:
        return [DynamicAttr.from_dict(k, v) for k, v in d.items() if k.startswith('dynamic_')]

    @classmethod
    def to_attributes(cls, dyn_attrs: List['DynamicAttr']) -> Dict[str, Dict[str, Any]]:
        return {f'dynamic_{d.field_id}': d.to_dict() for d in dyn_attrs}

    @classmethod
    def from_dict(cls, key: str, d: Dict[str, Any]) -> 'DynamicAttr':
        if key.startswith('dynamic_'):
            _, field_id = key.split('_', maxsplit=1)
            return DynamicAttr(field_id=int(field_id), label=d['label'], value=d['value'])
        else:
            raise ValueError(f"dynamic attribute '{key}' does not start with 'dynamic_'")

    def to_dict(self) -> Dict[str, Any]:
        return {'label': self.label, 'value': self.value}

    def clone(self, new_value: Optional[Any] = None):
        return DynamicAttr(field_id=self.field_id, label=self.label,
                           value=self.value if new_value is None else new_value)


@total_ordering
class PersonioResource:

    _api_type_name: str = None
    """the name of this resource type in the Personio API"""
    _field_mapping_list: List[FieldMapping] = []
    """all known API fields and their type definitions that are mapped to this PersonioResource"""
    __field_mapping: Dict[str, FieldMapping] = None
    """see ``_field_mapping()``"""
    __label_mapping: Dict[str, str] = None
    """see ``_label_mapping()``"""
    __namedtuple: Type[tuple] = None
    """see ``_namedtuple()``"""
    _flat_dict = False
    """set this to True, if this class has a flat dictionary representation in the Personio API"""

    def __init__(self, client: 'Personio' = None, **kwargs):
        super().__init__()
        self._client = client

    @classmethod
    def _field_mapping(cls) -> Dict[str, FieldMapping]:
        # the field mapping as dictionary
        if cls.__field_mapping is None:
            cls.__field_mapping = {fm.api_field: fm for fm in cls._field_mapping_list}
        return cls.__field_mapping

    @classmethod
    def _label_mapping(cls) -> Dict[str, str]:
        # mapping from api field name to pretty label name
        if cls.__label_mapping is None:
            cls.__label_mapping = {}
        return cls.__label_mapping

    @classmethod
    def from_dict(cls, d: Dict[str, Any], client: 'Personio' = None) -> '__class__':
        """
        Create an instance of this PersonioResource from the specified dictionary data,
        which is parsed version of the json data from the Personio API.

        :param d: create an instance from this data
        :param client: the Personio API client (optional). Used to provide additional operations
               on this resource, when available (e.g. request more data or write changes back
               to Personio)
        :return: a new instance of this class based on the provided data
        """
        # handle 'type' & 'attributes', if available
        if 'type' in d and 'attributes' in d:
            cls._check_api_type(d)
            d = d['attributes']
        # map the dictionary contents to the constructor's parameter names
        kwargs = cls._map_fields(d, client)
        return cls(client=client, **kwargs)

    def to_dict(self, nested=False) -> Dict[str, Any]:
        """
        Convert this PersonioResource to a dictionary that has the same structure as the
        json data from the Personio API.

        :param nested: indicate that this resource is part of a nested dictionary
               (Personio resources can have a different serialization when they are part of
               a nested dictionary...)
        :return: the Personio resource as dictionary (same structure as in the Personio API)
        """
        d = {}
        for mapping in self._field_mapping_list:
            value = getattr(self, mapping.class_field)
            if value is not None:
                d[mapping.api_field] = mapping.serialize(value)
        return d

    @classmethod
    def _check_api_type(cls, d: Dict[str, Any]):
        api_type_name = d['type']
        if api_type_name != cls._api_type_name:
            log_once(
                logging.WARNING,
                f"Unexpected API type '{api_type_name}' for class {cls.__name__}, "
                f"expected '{cls._api_type_name}' instead")

    @classmethod
    def _namedtuple(cls) -> Type[Tuple]:
        if cls.__namedtuple is None:
            fields = [m.class_field for m in cls._field_mapping_list] + ['dynamic', 'class_name']
            cls.__namedtuple = namedtuple(f'{cls.__name__}Tuple', fields)
        return cls.__namedtuple

    def to_tuple(self) -> Tuple:
        values = ([getattr(self, m.class_field) for m in self._field_mapping_list] +
                  [getattr(self, 'dynamic'), str(self.__class__)])
        return self._namedtuple()(*values)

    @classmethod
    def _map_fields(cls, d: Dict[str, Dict[str, Any]], client: 'Personio' = None) -> Dict[str, Any]:
        kwargs = {}
        field_mapping_dict = cls._field_mapping()
        for key, value in d.items():
            if key in field_mapping_dict:
                field_mapping = field_mapping_dict[key]
                if not cls._is_empty(value):
                    value = field_mapping.deserialize(value, client=client)
                kwargs[field_mapping.class_field] = value
            else:
                log_once(logging.WARNING, f"unexpected field '{key}' in class {cls.__name__}")
        return kwargs

    @classmethod
    def _is_empty(cls, value: Any):
        # determine if this Personio API value is "empty".
        # empty values are: None, "", []
        # not empty values are: 0, False, "foo", [1,2,3], 42
        return value is None or value == "" or value == []

    def __hash__(self):
        return hash(json.dumps(self.to_tuple(), sort_keys=True, default=str))

    def __eq__(self, other):
        if isinstance(other, PersonioResource):
            return self.to_tuple() == other.to_tuple()
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, PersonioResource):
            return self.to_tuple() < other.to_tuple()
        else:
            return False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.__dict__}"

    def __str__(self) -> str:
        fields = ', '.join(f'{k}={v}' for k, v in self.__dict__.items() if not k.startswith('_'))
        return f"{self.__class__.__name__}({fields})"


PersonioResourceType = TypeVar('PersonioResourceType', bound=PersonioResource)


class WritablePersonioResource(PersonioResource):

    _can_create = True
    _can_update = True
    _can_delete = True

    def __init__(self, client: 'Personio' = None, dynamic: List['DynamicAttr'] = None,
                 dynamic_fields: List[DynamicMapping] = None, **kwargs):
        super().__init__(client, **kwargs)
        self.dynamic_fields = dynamic_fields
        self.dynamic_raw: Dict[int, DynamicAttr] = {d.field_id: d for d in dynamic or []}
        self.dynamic = self._map_dynamic_values(dynamic, dynamic_fields, client)

    @classmethod
    def _map_dynamic_values(
            cls, dynamic_raw: List['DynamicAttr'], dynamic_fields: List[DynamicMapping] = None,
            client: 'Personio' = None) -> Dict[str, Any]:
        dynamic = {}
        if not dynamic_raw or not dynamic_fields:
            return dynamic
        dynamic_mapping_dict = {dm.field_id: dm for dm in dynamic_fields or []}
        for dyn in dynamic_raw:
            if dyn.field_id in dynamic_mapping_dict:
                # we have a dynamic field mapping -> parse the value
                dm: DynamicMapping = dynamic_mapping_dict[dyn.field_id]
                field_mapping = dm.get_field_mapping()
                value = dyn.value
                if not cls._is_empty(value):
                    value = field_mapping.deserialize(value, client=client)
                dynamic[field_mapping.class_field] = value
        return dynamic

    @classmethod
    def from_dict(cls, d: Dict[str, Any], client: 'Personio' = None,
                  dynamic_fields: List[DynamicMapping] = None) -> '__class__':
        cls._check_api_type(d)
        kwargs = cls._map_fields(d['attributes'], client)
        if 'id' in d:
            kwargs['id_'] = d['id']
        dynamic_fields = dynamic_fields or (client.dynamic_fields if client else None)
        return cls(client=client, dynamic_fields=dynamic_fields, **kwargs)

    def to_dict(self, nested=False) -> Dict[str, Any]:
        # we prefer typed values from the dynamic dict over the raw values
        # (because they might have been changed by the user)
        attr = super().to_dict(nested)
        dynamic_mapping_dict = {dyn.field_id: dyn for dyn in self.dynamic_fields or []}
        for dyn in self.dynamic_raw.values():
            if dyn.field_id in dynamic_mapping_dict:
                raw_value = dyn.value
                dm: DynamicMapping = dynamic_mapping_dict[dyn.field_id]
                rich_value = self.dynamic[dm.alias]
                if raw_value != rich_value:
                    field_mapping = dm.get_field_mapping()
                    serialized = field_mapping.serialize(rich_value)
                    if raw_value != serialized:
                        dyn = dyn.clone(new_value=serialized)
            attr[f'dynamic_{dyn.field_id}'] = dyn.to_dict()
        return {
            'type': self._api_type_name,
            'attributes': attr,
        }

    def create(self, client: 'Personio' = None):
        if self._can_create:
            client = self._check_client(client)
            return self._create(client)
        else:
            raise UnsupportedMethodError('create', self.__class__)

    def _create(self, client: 'Personio'):
        raise UnsupportedMethodError('create', self.__class__)

    def update(self, client: 'Personio' = None):
        if self._can_update:
            client = self._check_client(client)
            return self._update(client)
        else:
            raise UnsupportedMethodError('update', self.__class__)

    def _update(self, client: 'Personio'):
        UnsupportedMethodError('update', self.__class__)

    def delete(self, client: 'Personio' = None):
        if self._can_delete:
            client = self._check_client(client)
            return self._delete(client)
        else:
            raise UnsupportedMethodError('delete', self.__class__)

    def _delete(self, client: 'Personio'):
        UnsupportedMethodError('delete', self.__class__)

    def _check_client(self, client: 'Personio' = None) -> 'Personio':
        client = client or self._client
        if not client:
            raise PersonioError()
        if not client.authenticated:
            client.authenticate()
        return client


class LabeledAttributesMixin(PersonioResource):
    """
    Personio Resources that use the ``LabeledAttributesMixin`` expect data in a different format
    than the regular key-value pattern. Example::

        "first_name": {
          "label": "First name",
          "value": "Richard"
        }

    Instead of ``"first_name": "Richard"`` we get a dictionary where the label of the field and
    its value are attributes of another dictionary.

    This format is currently used by ``Employee`` and ``ShortEmployee`` and was probably chosen
    because Personio allows to specify custom fields for employees with custom label names.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_dict(self, nested=False) -> Dict[str, Any]:
        d = {}
        label_mapping = self._label_mapping()
        for mapping in self._field_mapping_list:
            value = getattr(self, mapping.class_field)
            if value is not None:
                label = label_mapping.get(mapping.api_field)
                d[mapping.api_field] = {'label': label, 'value': mapping.serialize(value)}
        return d

    @classmethod
    def _map_fields(cls, d: Dict[str, Dict[str, Any]], client: 'Personio' = None) -> Dict[str, Any]:
        kwargs = {}
        dynamic = []
        field_mapping_dict = cls._field_mapping()
        label_mapping = cls._label_mapping()
        for key, data in d.items():
            label_mapping[key] = data['label']
            if key in field_mapping_dict:
                field_mapping = field_mapping_dict[key]
                value = data['value']
                if not cls._is_empty(value):
                    value = field_mapping.deserialize(value, client=client)
                kwargs[field_mapping.class_field] = value
            elif key.startswith('dynamic_'):
                dyn = DynamicAttr.from_dict(key, data)
                dynamic.append(dyn)
            else:
                log_once(logging.WARNING, f"unexpected field '{key}' in class {cls.__name__}")
        if dynamic:
            kwargs['dynamic'] = dynamic
        return kwargs


class AbsenceEntitlement(PersonioResource):

    _api_type_name = "TimeOffType"
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
        NumericFieldMapping('entitlement', 'entitlement', float),
    ]

    def __init__(self, id_: int = None, name: str = None, entitlement: float = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name
        self.entitlement = entitlement


class AbsenceType(PersonioResource):

    _api_type_name = "TimeOffType"
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
        FieldMapping('category', 'category', str)
    ]

    def __init__(self, id_: int = None, name: str = None, category: str = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name
        self.category = category

    def to_dict(self, nested=False) -> Dict[str, Any]:
        if nested:
            return super().to_dict()
        else:
            return {
                'type': self._api_type_name,
                'attributes': super().to_dict(),
            }


class Certificate(PersonioResource):

    _field_mapping_list = [
        FieldMapping('status', 'status', str),
    ]
    _flat_dict = True

    def __init__(self, status: str = None, **kwargs):
        super().__init__(**kwargs)
        self.status = status


class CostCenter(PersonioResource):

    _api_type_name = 'CostCenter'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
        NumericFieldMapping('percentage', 'percentage', float),
    ]

    def __init__(self, id_: int = None, name: str = None, percentage: float = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name
        self.percentage = percentage


class Department(PersonioResource):

    _api_type_name = 'Department'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
    ]

    def __init__(self, id_: int = None, name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name


class HolidayCalendar(PersonioResource):

    _api_type_name = 'HolidayCalendar'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
        FieldMapping('country', 'country', str),
        FieldMapping('state', 'state', str),
    ]

    def __init__(self, id_: int = None, name: str = None, country: str = None,
                 state: str = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name
        self.country = country
        self.state = state


class Office(PersonioResource):

    _api_type_name = 'Office'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
    ]

    def __init__(self, id_: int = None, name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name


class ShortEmployee(LabeledAttributesMixin):

    _api_type_name = "Employee"
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('first_name', 'first_name', str),
        FieldMapping('last_name', 'last_name', str),
        FieldMapping('email', 'email', str),
    ]

    def __init__(self, client: 'Personio' = None, id_: int = None, first_name: str = None,
                 last_name: str = None, email: str = None, **kwargs):
        super().__init__(**kwargs)
        self._client = client
        self.id_ = id_
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def resolve(self, client: 'Personio' = None) -> 'Employee':
        client = client or self._client
        if client:
            return client.get_employee(self.id_)
        else:
            raise PersonioError(
                f"no Personio client is is available in this {self.__class__.__name__} instance "
                f"to make a request for the full employee profile of "
                f"{self.first_name} {self.last_name} ({self.id_})")


class Team(PersonioResource):

    _api_type_name = 'Team'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
    ]

    def __init__(self, id_: int = None, name: str = None, **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name


class WorkSchedule(PersonioResource):

    _api_type_name = 'WorkSchedule'
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('name', 'name', str),
        DateFieldMapping('valid_from', 'valid_from'),
        DurationFieldMapping('monday', 'monday'),
        DurationFieldMapping('tuesday', 'tuesday'),
        DurationFieldMapping('wednesday', 'wednesday'),
        DurationFieldMapping('thursday', 'thursday'),
        DurationFieldMapping('friday', 'friday'),
        DurationFieldMapping('saturday', 'saturday'),
        DurationFieldMapping('sunday', 'sunday'),
    ]

    def __init__(self,
                 id_: int = None,
                 name: str = None,
                 valid_from: datetime = None,
                 monday: timedelta = None,
                 tuesday: timedelta = None,
                 wednesday: timedelta = None,
                 thursday: timedelta = None,
                 friday: timedelta = None,
                 saturday: timedelta = None,
                 sunday: timedelta = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.id_ = id_
        self.name = name
        self.valid_from = valid_from
        self.monday = monday
        self.tuesday = tuesday
        self.wednesday = wednesday
        self.thursday = thursday
        self.friday = friday
        self.saturday = saturday
        self.sunday = sunday


class Absence(WritablePersonioResource):

    _api_type_name = "TimeOffPeriod"
    _can_update = False
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('status', 'status', str),
        FieldMapping('comment', 'comment', str),
        DateFieldMapping('start_date', 'start_date'),
        DateFieldMapping('end_date', 'end_date'),
        NumericFieldMapping('days_count', 'days_count', float),
        NumericFieldMapping('half_day_start', 'half_day_start', int),
        NumericFieldMapping('half_day_end', 'half_day_end', int),
        ObjectFieldMapping('time_off_type', 'time_off_type', AbsenceType),
        ObjectFieldMapping('employee', 'employee', ShortEmployee),
        FieldMapping('created_by', 'created_by', str),
        ObjectFieldMapping('certificate', 'certificate', Certificate),
        DateTimeFieldMapping('created_at', 'created_at'),
        DateTimeFieldMapping('updated_at', 'updated_at'),
    ]

    def __init__(self,
                 client: 'Personio' = None,
                 dynamic: Dict[str, Any] = None,
                 dynamic_raw: List['DynamicAttr'] = None,
                 id_: int = None,
                 status: str = None,
                 comment: str = None,
                 start_date: datetime = None,
                 end_date: datetime = None,
                 days_count: float = None,
                 half_day_start: bool = False,
                 half_day_end: bool = False,
                 time_off_type: AbsenceType = None,
                 employee: ShortEmployee = None,
                 created_by: str = None,
                 certificate: Certificate = None,
                 created_at: datetime = None,
                 updated_at: datetime = None,
                 **kwargs):
        super().__init__(client=client, dynamic=dynamic, dynamic_raw=dynamic_raw, **kwargs)
        self.id_ = id_
        self.status = status
        self.comment = comment
        self.start_date = start_date
        self.end_date = end_date
        self.days_count = days_count
        self.half_day_start = bool(half_day_start)
        self.half_day_end = bool(half_day_end)
        self.time_off_type = time_off_type
        self.employee = employee
        self.created_by = created_by
        self.certificate = certificate
        self.created_at = created_at
        self.updated_at = updated_at

    def _create(self, client: 'Personio' = None):
        return get_client(self, client).create_absence(self)

    def _delete(self, client: 'Personio' = None):
        return get_client(self, client).delete_absence(self)

    def to_body_params(self):
        data = {
            'employee_id': self.employee.id_,
            'time_off_type_id': self.time_off_type.id_,
            'start_date': self.start_date.strftime("%Y-%m-%d"),
            'end_date': self.end_date.strftime("%Y-%m-%d"),
            'half_day_start': self.half_day_start,
            'half_day_end': self.half_day_end
        }
        if self.comment is not None:
            data['comment'] = self.comment
        return data


class Attendance(WritablePersonioResource):

    _api_type_name = "AttendancePeriod"
    _field_mapping_list = [
        # note: the id is actually not in the attributes dict, but one level higher
        NumericFieldMapping('id', 'id_', int),
        NumericFieldMapping('employee', 'employee_id', int),
        DateFieldMapping('date', 'date'),
        DurationFieldMapping('start_time', 'start_time'),
        DurationFieldMapping('end_time', 'end_time'),
        NumericFieldMapping('break', 'break_duration', int),
        FieldMapping('comment', 'comment', str),
        BooleanFieldMapping('is_holiday', 'is_holiday'),
        BooleanFieldMapping('is_on_time_off', 'is_on_time_off'),
    ]

    def __init__(self,
                 client: 'Personio' = None,
                 dynamic: Dict[str, Any] = None,
                 dynamic_raw: List['DynamicAttr'] = None,
                 id_: int = None,
                 employee_id: int = None,
                 date: datetime = None,
                 start_time: str = None,
                 end_time: str = None,
                 break_duration: int = None,
                 comment: str = None,
                 is_holiday: bool = None,
                 is_on_time_off: bool = None,
                 **kwargs):
        super().__init__(client=client, dynamic=dynamic, dynamic_raw=dynamic_raw, **kwargs)
        self.id_ = id_
        self.employee_id = employee_id
        self.date = date
        self.start_time = start_time
        self.end_time = end_time
        self.break_duration = break_duration
        self.comment = comment
        self.is_holiday = is_holiday
        self.is_on_time_off = is_on_time_off

    def to_dict(self, nested=False) -> Dict[str, Any]:
        # yes, this is weird an unnecessary, but that's how the api works
        d = super().to_dict()
        d['id'] = self.id_
        del d['attributes']['id']
        return d

    def _create(self, client: 'Personio'):
        pass

    def _update(self, client: 'Personio'):
        pass

    def _delete(self, client: 'Personio'):
        pass


class Employee(WritablePersonioResource, LabeledAttributesMixin):

    _api_type_name = "Employee"
    _can_delete = False
    _field_mapping_list = [
        NumericFieldMapping('id', 'id_', int),
        FieldMapping('first_name', 'first_name', str),
        FieldMapping('last_name', 'last_name', str),
        FieldMapping('email', 'email', str),
        FieldMapping('gender', 'gender', str),
        FieldMapping('status', 'status', str),
        FieldMapping('position', 'position', str),
        ObjectFieldMapping('supervisor', 'supervisor', ShortEmployee),
        FieldMapping('employment_type', 'employment_type', str),
        FieldMapping('weekly_working_hours', 'weekly_working_hours', str),
        DateFieldMapping('hire_date', 'hire_date'),
        DateFieldMapping('contract_end_date', 'contract_end_date'),
        DateFieldMapping('termination_date', 'termination_date'),
        FieldMapping('termination_type', 'termination_type', str),
        FieldMapping('termination_reason', 'termination_reason', str),
        DateFieldMapping('probation_period_end', 'probation_period_end'),
        DateTimeFieldMapping('created_at', 'created_at'),
        DateTimeFieldMapping('last_modified_at', 'last_modified_at'),
        FieldMapping('subcompany', 'subcompany', str),
        ObjectFieldMapping('office', 'office', Office),
        ObjectFieldMapping('department', 'department', Department),
        ListFieldMapping(ObjectFieldMapping(
            'cost_centers', 'cost_centers', CostCenter)),
        NumericFieldMapping('fix_salary', 'fix_salary', float),
        FieldMapping('fix_salary_interval', 'fix_salary_interval', str),
        NumericFieldMapping('hourly_salary', 'hourly_salary', float),
        NumericFieldMapping('vacation_day_balance', 'vacation_day_balance', float),
        DateFieldMapping('last_working_day', 'last_working_day'),
        ObjectFieldMapping('holiday_calendar', 'holiday_calendar', HolidayCalendar),
        ObjectFieldMapping('work_schedule', 'work_schedule', WorkSchedule),
        ListFieldMapping(ObjectFieldMapping(
            'absence_entitlement', 'absence_entitlement', AbsenceEntitlement)),
        FieldMapping('profile_picture', 'profile_picture', str),
        ObjectFieldMapping('team', 'team', Team),
    ]

    def __init__(self,
                 client: 'Personio' = None,
                 dynamic: Dict[str, Any] = None,
                 dynamic_raw: List['DynamicAttr'] = None,
                 id_: int = None,
                 first_name: str = None,
                 last_name: str = None,
                 email: str = None,
                 gender: str = None,
                 status: str = None,
                 position: str = None,
                 supervisor: ShortEmployee = None,
                 employment_type: str = None,
                 weekly_working_hours: str = None,
                 hire_date: datetime = None,
                 contract_end_date: datetime = None,
                 termination_date: datetime = None,
                 termination_type: str = None,
                 termination_reason: str = None,
                 probation_period_end: datetime = None,
                 created_at: datetime = None,
                 last_modified_at: datetime = None,
                 subcompany: str = None,
                 office: Office = None,
                 department: Department = None,
                 cost_centers: List[CostCenter] = None,
                 holiday_calendar: HolidayCalendar = None,
                 absence_entitlement: List[AbsenceEntitlement] = None,
                 work_schedule: WorkSchedule = None,
                 fix_salary: float = None,
                 fix_salary_interval: str = None,
                 hourly_salary: float = None,
                 vacation_day_balance: float = None,
                 last_working_day: datetime = None,
                 profile_picture: str = None,
                 team: Team = None,
                 **kwargs):
        super().__init__(client=client, dynamic=dynamic, dynamic_raw=dynamic_raw, **kwargs)
        self.id_ = id_
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.gender = gender
        self.status = status
        self.position = position
        self.supervisor = supervisor
        self.employment_type = employment_type
        self.weekly_working_hours = weekly_working_hours
        self.hire_date = hire_date
        self.contract_end_date = contract_end_date
        self.termination_date = termination_date
        self.termination_type = termination_type
        self.termination_reason = termination_reason
        self.probation_period_end = probation_period_end
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.subcompany = subcompany
        self.office = office
        self.department = department
        self.cost_centers = cost_centers
        self.holiday_calendar = holiday_calendar
        self.absence_entitlement = absence_entitlement
        self.work_schedule = work_schedule
        self.fix_salary = fix_salary
        self.fix_salary_interval = fix_salary_interval
        self.hourly_salary = hourly_salary
        self.vacation_day_balance = vacation_day_balance
        self.last_working_day = last_working_day
        self.profile_picture = profile_picture
        self.team = team
        self._picture = None

    def _create(self, client: 'Personio' = None):
        pass

    def _update(self, client: 'Personio' = None):
        pass

    def picture(self, client: 'Personio' = None, width: int = None) -> bytes:
        if self._picture is None:
            client = get_client(self, client)
            self._picture = client.get_employee_picture(self, width=width)
        return self._picture

    def __str__(self):
        return f"{self.__class__.__name__}: {self.first_name} {self.last_name}, " \
               f"{self.position or 'position undefined'} ({self.id_})"


_unique_logs = set()


def log_once(level: int, message: str):
    if message not in _unique_logs:
        logger.log(level, message)
        _unique_logs.add(message)


def get_client(resource: PersonioResource, client: 'Personio' = None):
    if resource._client or client:
        return resource._client or client
    raise PersonioError(f"no Personio client reference is available, please provide it to "
                        f"your {type(resource).__name__} or as function parameter")



"""
Implementation of the Personio API functions
"""
import logging
import os
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from urllib.parse import urljoin

import requests
from requests import Response

from personio_py import Absence, AbsenceType, Attendance, DynamicMapping, Employee
from personio_py.errors import MissingCredentialsError, PersonioApiError, PersonioError
from personio_py.models import PersonioResource
from personio_py.search import SearchIndex

logger = logging.getLogger('personio_py')

PersonioResourceType = TypeVar('PersonioResourceType', bound=PersonioResource, covariant=True)


class Personio:
    """
    the Personio API client.

    :param base_url: use this custom base URL instead of the default https://api.personio.de/v1/
    :param client_id: the client id for API authentication
           (if not provided, defaults to the ``CLIENT_ID`` environment variable)
    :param client_secret: the client secret for API authentication
           (if not provided, defaults to the ``CLIENT_SECRET`` environment variable)
    :param dynamic_fields: definition of expected dynamic fields.
           List of :py:class:`DynamicMapping` tuples.
    """

    BASE_URL = "https://api.personio.de/v1/"
    """base URL of the Personio HTTP API"""

    def __init__(self, base_url: str = None, client_id: str = None, client_secret: str = None,
                 dynamic_fields: List[DynamicMapping] = None):
        self.base_url = base_url or self.BASE_URL
        self.client_id = client_id or os.getenv('CLIENT_ID')
        self.client_secret = client_secret or os.getenv('CLIENT_SECRET')
        self.headers = {'accept': 'application/json'}
        self.authenticated = False
        self.dynamic_fields = dynamic_fields
        self.search_index = SearchIndex(self)

    def authenticate(self):
        """
        Try to authenticate (using Personio's ``/auth`` endpoint) with the credentials
        (client ID and secret) that were provided to this instance.

        There should be no need to explicitly call this function, because you will be automatically
        authenticated when you make your first request.

        If the authentication was successful, the authentication token is stored in the
        headers dictionary (``self.headers``). This token will be sent with every request and
        it will be automatically rotated whenever necessary.
        If the authentication failed, a ``PersonioApiError`` will be raised.
        """
        if not (self.client_id and self.client_secret):
            raise MissingCredentialsError(
                "both client_id and client_secret must be provided in order to authenticate")
        url = urljoin(self.base_url, 'auth')
        logger.debug(f"authenticating to {url} with client_id {self.client_id}")
        params = {"client_id": self.client_id, "client_secret": self.client_secret}
        response = requests.request("POST", url, headers=self.headers, params=params)
        if response.ok:
            token = response.json()['data']['token']
            self.headers['Authorization'] = f"Bearer {token}"
            self.authenticated = True
        else:
            raise PersonioApiError.from_response(response)

    def request(self, path: str, method='GET', params: Dict[str, Any] = None,
                data: Dict[str, Any] = None, headers: Dict[str, str] = None,
                auth_rotation=True) -> Response:
        """
        Make a request against the Personio API.
        Returns the HTTP response, which might be successful or not.

        Rotation of authorization tokens is handled in this function. If no token is available
        (first request of this instance), the ``authenticate()`` function is called. On subsequent
        requests, the new authorization token is read from the response headers and replaces the
        previous one. Note that this behavior can be turned off by setting auth_rotation=False,
        which is required for some requests that do not return a new authorization token.

        :param path: the URL path for this request (relative to the Personio API base URL)
        :param method: the HTTP request method (default: GET)
        :param params: dictionary of URL parameters (optional)
        :param data: dictionary of data to send in the request body (optional)
        :param headers: additional request headers (authentication is handled separately)
        :param auth_rotation: set to True, if authentication keys should be rotated
               during this request (default: True)
        :return: the HTTP response (the caller is responsible for handling HTTP errors)
        """
        # check if we are already authenticated
        if not self.authenticated:
            self.authenticate()
        # define headers
        _headers = self.headers
        if headers:
            _headers.update(headers)
        # make the request
        url = urljoin(self.base_url, path)
        response = requests.request(method, url, headers=_headers, params=params, json=data)
        # re-new the authorization header
        authorization = response.headers.get('Authorization')
        if authorization:
            self.headers['Authorization'] = authorization
        elif auth_rotation:
            raise PersonioError("Missing Authorization Header in response")
        # return the response, let the caller handle any issues
        return response

    def request_json(self, path: str, method='GET', params: Dict[str, Any] = None,
                     data: Dict[str, Any] = None, auth_rotation=True) -> Dict[str, Any]:
        """
        Make a request against the Personio API, expecting a json response.
        Returns the parsed json response as dictionary. Will raise a PersonioApiError if the
        request fails.

        :param path: the URL path for this request (relative to the Personio API base URL)
        :param method: the HTTP request method (default: GET)
        :param params: dictionary of URL parameters (optional)
        :param data: dictionary of data to send in the request body (optional)
        :param auth_rotation: set to True, if authentication keys should be rotated
               during this request (default: True for json requests)
        :return: the parsed json response, when the request was successful, or a PersonioApiError
        """
        response = self.request(path, method, params, data, auth_rotation=auth_rotation)
        if response.ok:
            try:
                return response.json()
            except ValueError:
                raise PersonioError(f"Failed to parse response as json: {response.text}")
        else:
            raise PersonioApiError.from_response(response)

    def request_paginated(
            self, path: str, method='GET', params: Dict[str, Any] = None,
            data: Dict[str, Any] = None, auth_rotation=True, limit=200) -> Dict[str, Any]:
        """
        Make a request against the Personio API, expecting a json response that may be paginated,
        i.e. not all results might have been returned after the first request. Will continue
        to make requests until no more results are provided by the Personio API.
        Returns the parsed json response as dictionary. Will raise a PersonioApiError if the
        request fails.

        :param path: the URL path for this request (relative to the Personio API base URL)
        :param method: the HTTP request method (default: GET)
        :param params: dictionary of URL parameters (optional)
        :param data: dictionary of data to send in the request body (optional)
        :param auth_rotation: set to True, if authentication keys should be rotated
               during this request (default: True for json requests)
        :param limit: the max. number of items to return in response to a single request.
               A higher limit means fewer requests will be made (though there is an upper bound
               that is enforced on the server side)
        :return: the parsed json response, when the request was successful, or a PersonioApiError
        """
        # prepare the params dict (need limit and offset as parameters)
        if params is None:
            params = {}
        params['limit'] = limit
        params['offset'] = 1
        # continue making requests until no more data is returned
        data_acc = []
        while True:
            response = self.request_json(path, method, params, data, auth_rotation=auth_rotation)
            resp_data = response['data']
            if resp_data:
                data_acc.extend(resp_data)
                if response['metadata']['current_page'] == response['metadata']['total_pages']:
                    break
                else:
                    params['offset'] += 1
            else:
                break
        # return the accumulated data
        response['data'] = data_acc
        return response

    def request_image(self, path: str, method='GET', params: Dict[str, Any] = None,
                      auth_rotation=False) -> Optional[bytes]:
        """
        Request an image file (as png or jpg) from the Personio API.
        Returns the image as byte array, or None, if no image is available for this resource
        (HTTP status 404).
        When any other errors occur, a PersonioApiError is raised.

        :param path: the URL path for this request (relative to the Personio API base URL)
        :param method: the HTTP request method (default: GET)
        :param params: dictionary of URL parameters (optional)
        :param auth_rotation: set to True, if authentication keys should be rotated
               during this request (default: False for image requests)
        :return: the image (bytes) or None, if no image is available
        """
        headers = {'accept': 'image/png, image/jpeg'}
        response = self.request(path, method, params, headers=headers, auth_rotation=auth_rotation)
        if response.ok:
            # great, we have our image as png or jpg
            return response.content
        elif response.status_code == 404:
            # no image? how disappointing...
            return None
        else:
            # oh noes, something went terribly wrong!
            raise PersonioApiError.from_response(response)

    def get_employees(self) -> List[Employee]:
        """
        Get a list of all employee records in your account.

        :return: list of ``Employee`` instances
        """
        response = self.request_json('company/employees')
        employees = [Employee.from_dict(d, self) for d in response['data']]
        return employees

    def get_employee(self, employee_id: int) -> Employee:
        """
        Get a single employee with the specified ID.

        :param employee_id: the Personio ID of the employee to fetch
        :return: an ``Employee`` instance or a PersonioApiError, if the employee does not exist
        """
        response = self.request_json(f'company/employees/{employee_id}')
        employee = Employee.from_dict(response['data'], self)
        return employee

    def get_employee_picture(self, employee: Union[int, Employee], width: int = None) \
            -> Optional[bytes]:
        """
        Get the profile picture of the specified employee as image file
        (usually png or jpg).

        :param employee: get the picture of this employee or the employee with
               the specified Personio ID
        :param width: optionally scale the profile picture to this width.
               Defaults to the original width of the profile picture.
        :return: the profile picture as png or jpg file (bytes)
        """
        employee_id = employee.id_ if isinstance(employee, Employee) else int(employee)
        path = f'company/employees/{employee_id}/profile-picture'
        if width:
            path += f'/{width}'
        return self.request_image(path, auth_rotation=False)

    def create_employee(self, employee: Employee, refresh=True) -> Employee:
        """
        placeholder; not ready to be used
        """
        # TODO warn about limited selection of fields
        data = {
            'employee[email]': employee.email,
            'employee[first_name]': employee.first_name,
            'employee[last_name]': employee.last_name,
            'employee[gender]': employee.gender,
            'employee[position]': employee.position,
            'employee[department]': employee.department.name,
            'employee[hire_date]': employee.hire_date.isoformat()[:10],
            'employee[weekly_hours]': employee.weekly_working_hours,
        }
        response = self.request_json('company/employees', method='POST', data=data)
        employee.id_ = response['data']['id']
        if refresh:
            return self.get_employee(employee.id_)
        else:
            return employee

    def update_employee(self, employee: Employee):
        """
        placeholder; not ready to be used
        """
        raise NotImplementedError()

    def get_attendances(self, employees: Union[int, List[int], Employee, List[Employee]],
                        start_date: datetime = None, end_date: datetime = None) -> List[Attendance]:
        """
        Get a list of all attendance records for the employees with the specified IDs

        Note that internally, multiple requests may be made by this function due to limitations
        of the Personio API: Only a limited number of records can be retrieved in a single request
        and only a limited number of employee IDs can be passed as URL parameters. The results
        are still presented as a single list, no matter how many requests are made.

        :param employees: a single employee or a list of employee objects or IDs.
               Attendance records for all matching employees will be retrieved.
        :param start_date: only return attendance records from this date (inclusive, optional)
        :param end_date: only return attendance records up to this date (inclusive, optional)
        :return: list of ``Attendance`` records for the specified employees
        """
        return self._get_employee_metadata(
            'company/attendances', Attendance, employees, start_date, end_date)

    def create_attendances(self, attendances: List[Attendance]):
        """
        placeholder; not ready to be used
        """
        # attendances can be created individually, but here you can push a huge bunch of items
        # in a single request, which can be significantly faster
        raise NotImplementedError()

    def update_attendance(self, attendance_id: int):
        """
        placeholder; not ready to be used
        """
        raise NotImplementedError()

    def delete_attendance(self, attendance_id: int):
        """
        placeholder; not ready to be used
        """
        raise NotImplementedError()

    def get_absence_types(self) -> List[AbsenceType]:
        """
        Get a list of all available absence types, e.g. "paid vacation" or "parental leave".

        The absence types are used to classify the absences of employees
        (see ``get_absences`` to get a list of all absences for the employees).
        Each ``Absence`` also contains the ``AbsenceType`` for this instance; the purpose
        of this function is to provide you with a list of all possible options that can show up.
        """
        response = self.request_json('company/time-off-types')
        absence_types = [AbsenceType.from_dict(d, self) for d in response['data']]
        return absence_types

    def get_absences(self, employees: Union[int, List[int], Employee, List[Employee]],
                     start_date: datetime = None, end_date: datetime = None) -> List[Absence]:
        """
        Get a list of all absence records for the employees with the specified IDs.

        Note that internally, multiple requests may be made by this function due to limitations
        of the Personio API: Only a limited number of records can be retrieved in a single request
        and only a limited number of employee IDs can be passed as URL parameters. The results
        are still presented as a single list, no matter how many requests are made.

        :param employees: a single employee or a list of employee objects or IDs.
               Absence records for all matching employees will be retrieved.
        :param start_date: only return absence records from this date (inclusive, optional)
        :param end_date: only return absence records up to this date (inclusive, optional)
        :return: list of ``Absence`` records for the specified employees
        """
        return self._get_employee_metadata(
            'company/time-offs', Absence, employees, start_date, end_date)

    def get_absence(self, absence: Union[Absence, int]) -> Absence:
        """
        Get an absence record from a given id.

        :param absence: The absence id to fetch.
        """
        if isinstance(absence, int):
            response = self.request_json(f'company/time-offs/{absence}')
            return Absence.from_dict(response['data'], self)
        else:
            if absence.id_:
                return self.get_absence(absence.id_)
            else:
                self.__add_remote_absence_id(absence)
                return self.get_absence(absence.id_)

    def create_absence(self, absence: Absence) -> Absence:
        """
        Creates an absence record on the Personio servers

        :param absence: The absence object to be created
        :raises PersonioError: If the absence could not be created on the Personio servers
        """
        data = absence.to_body_params()
        response = self.request_json('company/time-offs', method='POST', data=data)
        if response['success']:
            absence.id_ = response['data']['attributes']['id']
            return absence
        raise PersonioError("Could not create absence")

    def delete_absence(self, absence: Union[Absence, int]):
        """
        Delete an existing record

        An absence id is required.

        :param absence: The Absence object holding
            the new data or an absence record id to delete.
        :raises ValueError: If a query is required but not allowed
            or the query does not provide exactly one result.
        """
        if isinstance(absence, int):
            response = self.request_json(path=f'company/time-offs/{absence}', method='DELETE')
            return response['success']
        elif isinstance(absence, Absence):
            if absence.id_ is not None:
                return self.delete_absence(absence.id_)
            else:
                raise ValueError("Only an absence with an absence id can be deleted.")
        else:
            raise ValueError("absence must be an Absence object or an integer")

    def search(self, query: str, active_only=True) -> List[Employee]:
        """
        Execute a search on the search index.

        If the index does not exist, or has been invalidated or is expired, the full list
        of employees will be requested from the API.

        During the search we perform a case insensitive match on the keywords in the search index.
        All tokens of the query will be matched individually. Tokens are separated by whitespace.
        A full match (i.e. all tokens match the keywords in order) is preferred over
        a partial match (only one or more tokens match).

        For more details, please refer to :class:`SearchIndex`.

        :param query: the query string
        :param active_only: exclude inactive employees from the results (default: yes)
        :return: the list of employees that matches the search query
        """
        return self.search_index.search(query, active_only=active_only)

    def search_first(self, query: str, active_only=True) -> Optional[Employee]:
        """
        Execute a search on the search index and return the first result (if there is one) or None.

        This is basically the "I'm Feeling Lucky" button.
        For details about the search function, please refer to :class:`SearchIndex`.

        :param query: the query string
        :param active_only: exclude inactive employees from the results (default: yes)
        :return: the first search result or None, if there were no results
        """
        return self.search_index.search_first(query, active_only=active_only)

    def invalidate_index(self):
        """
        Invalidates the search index. New data will be requested on the next search.
        """
        self.search_index.invalidate()

    def _get_employee_metadata(
            self, path: str, resource_cls: Type[PersonioResourceType],
            employees: Union[int, List[int], Employee, List[Employee]], start_date: datetime = None,
            end_date: datetime = None) -> List[PersonioResourceType]:
        # resolve params to match API requirements
        employees, start_date, end_date = self._normalize_timeframe_params(
            employees, start_date, end_date)
        params = {
            "start_date": start_date.isoformat()[:10],
            "end_date": end_date.isoformat()[:10],
        }
        # request in batches of up to 50 employees (keeps URL length well below 2000 chars)
        data_acc = []
        for i in range(0, len(employees), 50):
            params["employees[]"] = employees[i:i + 50]
            response = self.request_paginated(path, params=params)
            data_acc.extend(response['data'])
        # create objects from accumulated API responses
        parsed_data = [resource_cls.from_dict(d, self) for d in data_acc]
        return parsed_data

    @classmethod
    def _normalize_timeframe_params(
            cls, employees: Union[int, List[int], Employee, List[Employee]],
            start_date: datetime = None, end_date: datetime = None) \
            -> Tuple[List[int], datetime, datetime]:
        """
        Whenever we need a list of employee IDs, a start date and an end date, this function comes
        in handy:

        * wraps a single employee ID into a list
        * sets the start date way into the past, if it was not provided
        * sets the end date way into the future, if it was not provided

        :param employees: a single employee or a list of employees (employee objects or just IDs)
        :param start_date: a start date (optional)
        :param end_date: an end date (optional)
        :return: a tuple of (list of employee IDs, start date, end date), no None values.
        """
        if not employees:
            raise ValueError("need at least one employee ID, got nothing")
        if start_date is None:
            start_date = datetime(1900, 1, 1)
        if end_date is None:
            end_date = datetime(datetime.now().year + 10, 1, 1)
        if not isinstance(employees, list):
            employees = [employees]
        employee_ids = [(e.id_ if isinstance(e, Employee) else e) for e in employees]
        return employee_ids, start_date, end_date

    def __add_remote_absence_id(self, absence: Absence) -> Absence:
        """
        Queries the API for an absence record matching
        the given Absence object and adds the remote id.

        :param absence: The absence object to be updated
        :return: The absence object with the absence_id set
        """
        if absence.employee is None:
            raise ValueError("For a remote query an employee_id is required")
        if absence.start_date is None:
            raise ValueError("For a remote query a start date is required")
        if absence.end_date is None:
            raise ValueError("For a remote query an end date is required")
        matching_remote_absences = self.get_absences(employees=[absence.employee.id_],
                                                     start_date=absence.start_date,
                                                     end_date=absence.end_date)
        if len(matching_remote_absences) == 0:
            raise PersonioError("The absence to patch was not found")
        elif len(matching_remote_absences) > 1:
            raise PersonioError("More than one absence found.")
        absence.id_ = matching_remote_absences[0].id_
        return absence



import logging
import time
from typing import Dict, List, Optional, TYPE_CHECKING

from personio_py import Employee

if TYPE_CHECKING:
    from personio_py import Personio

logger = logging.getLogger('personio_py')


class SearchIndex:
    """
    The search index is an implementation of a basic local search function for employees
    that was created because the Personio API lacks a native search function.

    This means that data for all employees will be downloaded from the API and indexed in memory
    before a search can be executed. On subsequent searches, the index will be reused and no more
    data is requested until the index is invalidated or becomes too old.

    To search for an employee, create an instance of the search index and submit a search query:

        index = SearchIndex(personio)
        results = index.search("Smith")

    The index requires a `Personio` client instance to request the list of employees from the API.
    Then we can search for all employees with the name "Smith".

    On the next search, no data will be requested from the API and the search will be a lot faster:

        whatshisname = index.search("john jim james")

    In the result list, we prefer exact matches, but each employee that has at least one token
    match will become part of the results. Also note that queries are case-insensitive.

    If we know exactly who we want and we're fairly certain that there is only one employee with
    that name, we can use `search_first` to get that employee directly

        zack = index.search("zacharias smith")

    If you made some changes or know that something must have changed on the server side, you can
    invalidate the index and make sure that the next search will happen on a fresh index:

        index.invalidate()

    Before the next search, the full list of employees will be requested from the API again and
    the search index will be rebuilt.

    :param client: the Personio API client (to request employee data)
    :param index_timeout: when the
    """

    DEFAULT_INDEX_TIMEOUT = 6 * 60 * 60
    """the default timeout for the search index (6 hours)"""

    def __init__(self, client: 'Personio', index_timeout: int = DEFAULT_INDEX_TIMEOUT):
        super().__init__()
        self.client = client
        self.index_timeout = index_timeout
        self.index: Optional[Dict[Employee, str]] = None
        self.last_update = 0.0
        self.valid = False

    def search(self, query: str, active_only=True) -> List[Employee]:
        """
        Execute a search on the search index.

        If the index does not exist, or has been invalidated or is expired, the full list
        of employees will be requested from the API.

        During the search we perform a case insensitive match on the keywords in the search index.
        All tokens of the query will be matched individually. Tokens are separated by whitespace.
        A full match (i.e. all tokens match the keywords in order) is preferred over
        a partial match (only one or more tokens match).

        :param query: the query string
        :param active_only: exclude inactive employees from the results (default: yes)
        :return: the list of employees that matches the search query
        """
        # update the search index if necessary
        self._update_on_demand()
        # prepare data
        query_norm = query.lower()
        query_tokens = [t.lower() for t in query.split()]
        full_match = []
        partial_match = []
        # run the actual search
        for employee, text in self.index.items():
            if active_only and employee.status == 'inactive':
                continue
            if query_norm in text:
                full_match.append(employee)
            else:
                for token in query_tokens:
                    if token in text:
                        partial_match.append(employee)
        # full matches first, then partial matches
        return full_match + partial_match

    def search_first(self, query: str, active_only=True) -> Optional[Employee]:
        """
        Execute a search on the search index and return the first result (if there is one) or None.

        This is basically the "I'm Feeling Lucky" button.
        For details about the search function, please refer to :func:`search`.

        :param query: the query string
        :param active_only: exclude inactive employees from the results (default: yes)
        :return: the first search result or None, if there were no results
        """
        results = self.search(query, active_only=active_only)
        if results:
            return results[0]

    def invalidate(self):
        """
        Invalidates the search index. New data will be requested on the next search.
        """
        self.valid = False

    def _update_on_demand(self):
        """
        Checks if an update is required and updates the index with new data from the API,
        if this is the case.
        """
        if not self.index:
            logger.debug("creating search index for the first time")
            self._update()
        elif not self.valid:
            logger.debug("updating search index because it was invalidated")
            self._update()
        elif time.time() > self.last_update + self.index_timeout:
            logger.debug(f"updating search index because it has not been updated "
                         f"for more than {self.index_timeout} seconds")
            self._update()

    def _update(self):
        """
        Requests a fresh list of employees and builds a new search index.
        """
        employees = self.client.get_employees()
        self.index = self._build_index(employees)
        self.last_update = time.time()
        self.valid = True

    @classmethod
    def _build_index(cls, employees: List[Employee]) -> Dict[Employee, str]:
        """
        Build a search index from the specified list of employees

        :param employees: the employees to index
        :return: the search index
        """
        return {e: cls._get_keywords(e) for e in employees}

    @classmethod
    def _get_keywords(cls, e: Employee) -> str:
        """
        Specifies the keywords that are to be stored for this employee as part of the search index.
        The keywords consist of all string fields in the employee object that someone might
        reasonably search for: name, position, email, department, team, etc.
        We intentionally avoid to add references to other persons, e.g. the supervisor, as we only
        have a very basic ranking and this may lead to inaccurate search results.
        the keywords are stored as a single lowercase string with tokens separated by whitespace.

        :param e: we calculate the keywords for this employee
        :return: the keywords for this employee as normalized string
                 with tokens separated by whitespace
        """
        keywords = [
            e.first_name,
            e.last_name,
            e.position,
            e.email,
            e.status,
            e.subcompany
        ]
        if e.office:
            keywords.append(e.office.name)
        if e.department:
            keywords.append(e.department.name)
        if e.team:
            keywords.append(e.team.name)
        keyword_string = ' '.join(t for t in keywords if t).lower()
        return keyword_string



import os
import shutil
import subprocess
import sys
from distutils.cmd import Command
from runpy import run_path

from setuptools import find_packages, setup

# read the program version from version.py (without loading the module)
__version__ = run_path('src/personio_py/version.py')['__version__']


def read(fname):
    """Utility function to read the README file."""
    return open(os.path.join(os.path.dirname(__file__), fname), 'r', encoding='utf-8').read()


class DistCommand(Command):

    description = "build the distribution packages (in the 'dist' folder)"
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        if os.path.exists('build'):
            shutil.rmtree('build')
        subprocess.run(["python", "setup.py", "sdist", "bdist_wheel"])


class TestCommand(Command):

    description = "run all tests with pytest"
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        sys.path.append('src')
        import pytest
        return pytest.main(['tests', '--no-cov'])


class TestCovCommand(Command):

    description = "run all tests with pytest and write a test coverage report"
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        sys.path.append('src')
        params = "tests --doctest-modules --junitxml=junit/test-results.xml " \
                 "--cov=src --cov-report=xml --cov-report=html"
        import pytest
        return pytest.main(params.split(' '))


setup(
    name="personio-py",
    version=__version__,
    author="Sebastian Straub",
    author_email="sebastian.straub@alexanderthamm.com",
    description="a lightweight Personio API client",
    url="https://github.com/at-gmbh/personio-py",
    project_urls={
        'Documentation': 'https://at-gmbh.github.io/personio-py/',
        'Source': 'https://github.com/at-gmbh/personio-py',
        'Tracker': 'https://github.com/at-gmbh/personio-py/issues',
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={'personio_py': ['res/*']},
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    install_requires=[
        'requests>=2.21.0,<3.0.0',
    ],
    tests_require=[
        'pytest',
        'pytest-cov',
        'pre-commit',
    ],
    cmdclass={
        'dist': DistCommand,
        'test': TestCommand,
        'testcov': TestCovCommand,
    },
    platforms='any',
    python_requires='>=3.7',
    license='Apache 2.0',
    classifiers=[
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Typing :: Typed',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



"""
personio-py

a lightweight Personio API client
"""

__title__ = "personio-py"
__copyright__ = "© 2020 Alexander Thamm GmbH"

from .version import __version__
from .errors import (
    PersonioError,
    MissingCredentialsError,
    PersonioApiError,
    UnsupportedMethodError,
)
from .mapping import (
    DynamicMapping
)
from .models import (
    Absence,
    AbsenceEntitlement,
    AbsenceType,
    Attendance,
    CostCenter,
    Department,
    DynamicAttr,
    Employee,
    HolidayCalendar,
    Office,
    ShortEmployee,
    Team,
    WorkSchedule,
)
from personio_py.client import Personio
